@@ -1,14 +1,18 @@
 package Method::Generate::Accessor;
 
 use Moo::_strictures;
-use Moo::_Utils;
+use Moo::_Utils qw(_load_module _maybe_load_module _install_coderef);
 use Moo::Object ();
-our @ISA = qw(Moo::Object);
+BEGIN { our @ISA = qw(Moo::Object) }
 use Sub::Quote qw(quote_sub quoted_from_sub quotify);
 use Scalar::Util 'blessed';
+use Carp qw(croak);
+BEGIN { our @CARP_NOT = qw(Moo::_Utils) }
 use overload ();
-use Module::Runtime qw(use_module);
 BEGIN {
+  *_CAN_WEAKEN_READONLY = (
+    "$]" < 5.008_003 or $ENV{MOO_TEST_PRE_583}
+  ) ? sub(){0} : sub(){1};
   our $CAN_HAZ_XS =
     !$ENV{MOO_XS_DISABLE}
       &&
@@ -21,20 +25,30 @@ BEGIN {
     (eval { Class::XSAccessor->VERSION('1.17') })
   ;
 }
+BEGIN {
+  package
+    Method::Generate::Accessor::_Generated;
+  $Carp::Internal{+__PACKAGE__} = 1;
+}
 
 my $module_name_only = qr/\A$Module::Runtime::module_name_rx\z/;
 
 sub _die_overwrite
 {
   my ($pkg, $method, $type) = @_;
-  die "You cannot overwrite a locally defined method ($method) with "
+  croak "You cannot overwrite a locally defined method ($method) with "
     . ( $type || 'an accessor' );
 }
 
 sub generate_method {
   my ($self, $into, $name, $spec, $quote_opts) = @_;
+  $quote_opts = {
+    no_defer => 1,
+    package => 'Method::Generate::Accessor::_Generated',
+    %{ $quote_opts||{} },
+  };
   $spec->{allow_overwrite}++ if $name =~ s/^\+//;
-  die "Must have an is" unless my $is = $spec->{is};
+  croak "Must have an is" unless my $is = $spec->{is};
   if ($is eq 'ro') {
     $spec->{reader} = $name unless exists $spec->{reader};
   } elsif ($is eq 'rw') {
@@ -48,7 +62,7 @@ sub generate_method {
     $spec->{reader} = $name unless exists $spec->{reader};
     $spec->{writer} = "_set_${name}" unless exists $spec->{writer};
   } elsif ($is ne 'bare') {
-    die "Unknown is ${is}";
+    croak "Unknown is ${is}";
   }
   if (exists $spec->{builder}) {
     if(ref $spec->{builder}) {
@@ -58,7 +72,7 @@ sub generate_method {
       $spec->{builder} = 1;
     }
     $spec->{builder} = '_build_'.$name if ($spec->{builder}||0) eq 1;
-    die "Invalid builder for $into->$name - not a valid method name"
+    croak "Invalid builder for $into->$name - not a valid method name"
       if $spec->{builder} !~ $module_name_only;
   }
   if (($spec->{predicate}||0) eq 1) {
@@ -77,11 +91,11 @@ sub generate_method {
     } elsif (blessed $isa and $isa->can('coerce')) {
       $spec->{coerce} = sub { $isa->coerce(@_) };
     } else {
-      die "Invalid coercion for $into->$name - no appropriate type constraint";
+      croak "Invalid coercion for $into->$name - no appropriate type constraint";
     }
   }
 
-  for my $setting (qw( isa coerce )) {
+  foreach my $setting (qw( isa coerce )) {
     next if !exists $spec->{$setting};
     $self->_validate_codulatable($setting, $spec->{$setting}, "$into->$name");
   }
@@ -98,7 +112,7 @@ sub generate_method {
       $spec->{moosify} = [$spec->{moosify}];
     }
 
-    for my $spec (@{$spec->{moosify}}) {
+    foreach my $spec (@{$spec->{moosify}}) {
       $self->_validate_codulatable('moosify', $spec, "$into->$name");
     }
   }
@@ -106,7 +120,7 @@ sub generate_method {
   my %methods;
   if (my $reader = $spec->{reader}) {
     _die_overwrite($into, $reader, 'a reader')
-      if !$spec->{allow_overwrite} && *{_getglob("${into}::${reader}")}{CODE};
+      if !$spec->{allow_overwrite} && defined &{"${into}::${reader}"};
     if (our $CAN_HAZ_XS && $self->is_simple_get($name, $spec)) {
       $methods{$reader} = $self->_generate_xs(
         getters => $into, $reader, $name, $spec
@@ -115,15 +129,16 @@ sub generate_method {
       $self->{captures} = {};
       $methods{$reader} =
         quote_sub "${into}::${reader}"
-          => '    die "'.$reader.' is a read-only accessor" if @_ > 1;'."\n"
+          => '    Carp::croak("'.$reader.' is a read-only accessor") if @_ > 1;'."\n"
              .$self->_generate_get($name, $spec)
           => delete $self->{captures}
+          => $quote_opts
         ;
     }
   }
   if (my $accessor = $spec->{accessor}) {
     _die_overwrite($into, $accessor, 'an accessor')
-      if !$spec->{allow_overwrite} && *{_getglob("${into}::${accessor}")}{CODE};
+      if !$spec->{allow_overwrite} && defined &{"${into}::${accessor}"};
     if (
       our $CAN_HAZ_XS
       && $self->is_simple_get($name, $spec)
@@ -138,12 +153,13 @@ sub generate_method {
         quote_sub "${into}::${accessor}"
           => $self->_generate_getset($name, $spec)
           => delete $self->{captures}
+          => $quote_opts
         ;
     }
   }
   if (my $writer = $spec->{writer}) {
     _die_overwrite($into, $writer, 'a writer')
-      if !$spec->{allow_overwrite} && *{_getglob("${into}::${writer}")}{CODE};
+      if !$spec->{allow_overwrite} && defined &{"${into}::${writer}"};
     if (
       our $CAN_HAZ_XS
       && $self->is_simple_set($name, $spec)
@@ -157,20 +173,23 @@ sub generate_method {
         quote_sub "${into}::${writer}"
           => $self->_generate_set($name, $spec)
           => delete $self->{captures}
+          => $quote_opts
         ;
     }
   }
   if (my $pred = $spec->{predicate}) {
     _die_overwrite($into, $pred, 'a predicate')
-      if !$spec->{allow_overwrite} && *{_getglob("${into}::${pred}")}{CODE};
+      if !$spec->{allow_overwrite} && defined &{"${into}::${pred}"};
     if (our $CAN_HAZ_XS && our $CAN_HAZ_XS_PRED) {
       $methods{$pred} = $self->_generate_xs(
         exists_predicates => $into, $pred, $name, $spec
       );
     } else {
       $methods{$pred} =
-        quote_sub "${into}::${pred}" =>
-          '    '.$self->_generate_simple_has('$_[0]', $name, $spec)."\n"
+        quote_sub "${into}::${pred}"
+          => $self->_generate_simple_has('$_[0]', $name, $spec)."\n"
+          => {}
+          => $quote_opts
         ;
     }
   }
@@ -179,10 +198,12 @@ sub generate_method {
   }
   if (my $cl = $spec->{clearer}) {
     _die_overwrite($into, $cl, 'a clearer')
-      if !$spec->{allow_overwrite} && *{_getglob("${into}::${cl}")}{CODE};
+      if !$spec->{allow_overwrite} && defined &{"${into}::${cl}"};
     $methods{$cl} =
-      quote_sub "${into}::${cl}" =>
-        $self->_generate_simple_clear('$_[0]', $name, $spec)."\n"
+      quote_sub "${into}::${cl}"
+        => $self->_generate_simple_clear('$_[0]', $name, $spec)."\n"
+        => {}
+        => $quote_opts
       ;
   }
   if (my $hspec = $spec->{handles}) {
@@ -194,31 +215,36 @@ sub generate_method {
         map [ $_ => ref($hspec->{$_}) ? @{$hspec->{$_}} : $hspec->{$_} ],
           keys %$hspec;
       } elsif (!ref($hspec)) {
-        map [ $_ => $_ ], use_module('Moo::Role')->methods_provided_by(use_module($hspec))
+        require Moo::Role;
+        _load_module $hspec;
+        map [ $_ => $_ ], Moo::Role->methods_provided_by($hspec)
       } else {
-        die "You gave me a handles of ${hspec} and I have no idea why";
+        croak "You gave me a handles of ${hspec} and I have no idea why";
       }
     };
     foreach my $delegation_spec (@specs) {
       my ($proxy, $target, @args) = @$delegation_spec;
       _die_overwrite($into, $proxy, 'a delegation')
-        if !$spec->{allow_overwrite} && *{_getglob("${into}::${proxy}")}{CODE};
+        if !$spec->{allow_overwrite} && defined &{"${into}::${proxy}"};
       $self->{captures} = {};
       $methods{$proxy} =
-        quote_sub "${into}::${proxy}" =>
-          $self->_generate_delegation($asserter, $target, \@args),
-          delete $self->{captures}
+        quote_sub "${into}::${proxy}"
+          => $self->_generate_delegation($asserter, $target, \@args)
+          => delete $self->{captures}
+          => $quote_opts
         ;
     }
   }
   if (my $asserter = $spec->{asserter}) {
-    $self->{captures} = {};
-
-
+    _die_overwrite($into, $asserter, 'an asserter')
+      if !$spec->{allow_overwrite} && defined &{"${into}::${asserter}"};
+    local $self->{captures} = {};
     $methods{$asserter} =
-      quote_sub "${into}::${asserter}" =>
-        $self->_generate_asserter($name, $spec),
-        delete $self->{captures};
+      quote_sub "${into}::${asserter}"
+        => $self->_generate_asserter($name, $spec)
+        => delete $self->{captures}
+        => $quote_opts
+      ;
   }
   \%methods;
 }
@@ -241,6 +267,11 @@ sub is_simple_set {
   !grep $spec->{$_}, qw(coerce isa trigger weak_ref);
 }
 
+sub has_default {
+  my ($self, $name, $spec) = @_;
+  $spec->{builder} or exists $spec->{default} or (($spec->{is}||'') eq 'lazy');
+}
+
 sub has_eager_default {
   my ($self, $name, $spec) = @_;
   (!$spec->{lazy} and (exists $spec->{default} or $spec->{builder}));
@@ -337,36 +368,34 @@ sub _generate_simple_get {
 
 sub _generate_set {
   my ($self, $name, $spec) = @_;
+  my ($me, $source) = ('$_[0]', '$_[1]');
   if ($self->is_simple_set($name, $spec)) {
-    $self->_generate_simple_set('$_[0]', $name, $spec, '$_[1]');
-  } else {
-    my ($coerce, $trigger, $isa_check) = @{$spec}{qw(coerce trigger isa)};
-    my $value_store = '$_[0]';
-    my $code;
-    if ($coerce) {
-      $value_store = '$value';
-      $code = "do { my (\$self, \$value) = \@_;\n"
-        ."        \$value = "
-        .$self->_generate_coerce($name, $value_store, $coerce).";\n";
-    }
-    else {
-      $code = "do { my \$self = shift;\n";
-    }
-    if ($isa_check) {
-      $code .=
-        "        ".$self->_generate_isa_check($name, $value_store, $isa_check).";\n";
-    }
-    my $simple = $self->_generate_simple_set('$self', $name, $spec, $value_store);
-    if ($trigger) {
-      my $fire = $self->_generate_trigger($name, '$self', $value_store, $trigger);
-      $code .=
-        "        ".$simple.";\n        ".$fire.";\n"
-        ."        $value_store;\n";
-    } else {
-      $code .= "        ".$simple.";\n";
-    }
-    $code .= "      }";
-    $code;
+    return $self->_generate_simple_set($me, $name, $spec, $source);
+  }
+
+  my ($coerce, $trigger, $isa_check) = @{$spec}{qw(coerce trigger isa)};
+  if ($coerce) {
+    $source = $self->_generate_coerce($name, $source, $coerce);
+  }
+  if ($isa_check) {
+    'scalar do { my $value = '.$source.";\n"
+    .'  ('.$self->_generate_isa_check($name, '$value', $isa_check)."),\n"
+    .'  ('.$self->_generate_simple_set($me, $name, $spec, '$value')."),\n"
+    .($trigger
+      ? '('.$self->_generate_trigger($name, $me, '$value', $trigger)."),\n"
+      : '')
+    .'  ('.$self->_generate_simple_get($me, $name, $spec)."),\n"
+    ."}";
+  }
+  elsif ($trigger) {
+    my $set = $self->_generate_simple_set($me, $name, $spec, $source);
+    "scalar (\n"
+    . '  ('.$self->_generate_trigger($name, $me, "($set)", $trigger)."),\n"
+    . '  ('.$self->_generate_simple_get($me, $name, $spec)."),\n"
+    . ")";
+  }
+  else {
+    '('.$self->_generate_simple_set($me, $name, $spec, $source).')';
   }
 }
 
@@ -422,24 +451,18 @@ sub _wrap_attr_exception {
   .'    name     => '.quotify($name).",\n"
   .'    step     => '.quotify($step).",\n"
   ."  };\n"
-  .($want_return ? '  my $_return;'."\n" : '')
-  .'  my $_error;'."\n"
-  ."  {\n"
-  .'    my $_old_error = $@;'."\n"
-  ."    if (!eval {\n"
-  .'      $@ = $_old_error;'."\n"
-  .($want_return ? '      $_return ='."\n" : '')
-  .'      '.$code.";\n"
-  ."      1;\n"
-  ."    }) {\n"
-  .'      $_error = $@;'."\n"
-  .'      if (!ref $_error) {'."\n"
-  .'        $_error = '.$prefix.'.$_error;'."\n"
-  ."      }\n"
-  ."    }\n"
-  .'    $@ = $_old_error;'."\n"
-  ."  }\n"
-  .'  die $_error if $_error;'."\n"
+  .($want_return ? '  (my $_return),'."\n" : '')
+  .'  (my $_error), (my $_old_error = $@);'."\n"
+  ."  (eval {\n"
+  .'    ($@ = $_old_error),'."\n"
+  .'    ('
+  .($want_return ? '$_return ='."\n" : '')
+  .$code."),\n"
+  ."    1\n"
+  ."  } or\n"
+  .'    $_error = ref $@ ? $@ : '.$prefix.'.$@);'."\n"
+  .'  ($@ = $_old_error),'."\n"
+  .'  (defined $_error and die $_error);'."\n"
   .($want_return ? '  $_return;'."\n" : '')
   ."}\n"
 }
@@ -494,68 +517,45 @@ sub generate_populate_set {
 
 sub _generate_populate_set {
   my ($self, $me, $name, $spec, $source, $test, $init_arg) = @_;
-  if ($self->has_eager_default($name, $spec)) {
-    my $get_indent = ' ' x ($spec->{isa} ? 6 : 4);
-    my $get_default = $self->_generate_get_default(
-                        '$new', $name, $spec
-                      );
-    my $get_value =
-      defined($spec->{init_arg})
-        ? "(\n${get_indent}  ${test}\n"
-            ."${get_indent}   ? ${source}\n${get_indent}   : "
+
+  my $has_default = $self->has_eager_default($name, $spec);
+  if (!($has_default || $test)) {
+    return '';
+  }
+  if ($has_default) {
+    my $get_default = $self->_generate_get_default($me, $name, $spec);
+    $source =
+      $test
+        ? "(\n  ${test}\n"
+            ."   ? ${source}\n   : "
             .$get_default
-            ."\n${get_indent})"
+            .")"
         : $get_default;
-    if ($spec->{coerce}) {
-      $get_value = $self->_generate_coerce(
-        $name, $get_value,
-        $spec->{coerce}, $init_arg
-      )
-    }
-    ($spec->{isa}
-      ? "    {\n      my \$value = ".$get_value.";\n      "
-        .$self->_generate_isa_check(
-          $name, '$value', $spec->{isa}, $init_arg
-        ).";\n"
-        .'      '.$self->_generate_simple_set($me, $name, $spec, '$value').";\n"
-        ."    }\n"
-      : '    '.$self->_generate_simple_set($me, $name, $spec, $get_value).";\n"
+  }
+  if ($spec->{coerce}) {
+    $source = $self->_generate_coerce(
+      $name, $source,
+      $spec->{coerce}, $init_arg
     )
-    .($spec->{trigger}
-      ? '    '
-        .$self->_generate_trigger(
-          $name, $me, $self->_generate_simple_get($me, $name, $spec),
-          $spec->{trigger}
-        )." if ${test};\n"
-      : ''
-    );
-  } else {
-    "    if (${test}) {\n"
-      .($spec->{coerce}
-        ? "      $source = "
-          .$self->_generate_coerce(
-            $name, $source,
-            $spec->{coerce}, $init_arg
-          ).";\n"
-        : ""
-      )
-      .($spec->{isa}
-        ? "      "
-          .$self->_generate_isa_check(
-            $name, $source, $spec->{isa}, $init_arg
-          ).";\n"
-        : ""
-      )
-      ."      ".$self->_generate_simple_set($me, $name, $spec, $source).";\n"
-      .($spec->{trigger}
-        ? "      "
-          .$self->_generate_trigger(
-            $name, $me, $self->_generate_simple_get($me, $name, $spec),
-            $spec->{trigger}
-          ).";\n"
-        : ""
-      )
-      ."    }\n";
+  }
+  if ($spec->{isa}) {
+    $source = 'scalar do { my $value = '.$source.";\n"
+    .'  ('.$self->_generate_isa_check(
+        $name, '$value', $spec->{isa}, $init_arg
+      )."),\n"
+    ."  \$value\n"
+    ."}\n";
+  }
+  my $set = $self->_generate_simple_set($me, $name, $spec, $source);
+  my $trigger = $spec->{trigger} ? $self->_generate_trigger(
+    $name, $me, $self->_generate_simple_get($me, $name, $spec),
+    $spec->{trigger}
+  ) : undef;
+  if ($has_default) {
+    "($set)," . ($trigger ? "($test and $trigger)," : '')
+  }
+  else {
+    "($test and ($set)" . ($trigger ? ", ($trigger)" : '') . "),";
   }
 }
 
@@ -583,21 +583,23 @@ sub _generate_simple_set {
     #
     # but requires Internal functions and is just too damn crazy
     # so simply throw a better exception
-    my $weak_simple = "do { Scalar::Util::weaken(${simple}); no warnings 'void'; $get }";
-    Moo::_Utils::lt_5_8_3() ? <<"EOC" : $weak_simple;
-      eval { Scalar::Util::weaken($simple); 1 }
-        ? do { no warnings 'void'; $get }
-        : do {
-          if( \$@ =~ /Modification of a read-only value attempted/) {
-            require Carp;
-            Carp::croak( sprintf (
-              'Reference to readonly value in "%s" can not be weakened on Perl < 5.8.3',
-              $name_str,
-            ) );
-          } else {
-            die \$@;
+    my $weak_simple = _CAN_WEAKEN_READONLY
+      ? "do { Scalar::Util::weaken(${simple}); no warnings 'void'; $get }"
+      : <<"EOC"
+        ( eval { Scalar::Util::weaken($simple); 1 }
+          ? do { no warnings 'void'; $get }
+          : do {
+            if( \$@ =~ /Modification of a read-only value attempted/) {
+              require Carp;
+              Carp::croak( sprintf (
+                'Reference to readonly value in "%s" can not be weakened on Perl < 5.8.3',
+                $name_str,
+              ) );
+            } else {
+              die \$@;
+            }
           }
-        }
+        )
 EOC
   } else {
     $simple;
@@ -615,9 +617,8 @@ sub _generate_asserter {
 
   "do {\n"
    ."  my \$val = ".$self->_generate_get($name, $spec).";\n"
-   ."  unless (".$self->_generate_simple_has('$_[0]', $name, $spec).") {\n"
-   .qq!    die "Attempted to access '${name}' but it is not set";\n!
-   ."  }\n"
+   ."  ".$self->_generate_simple_has('$_[0]', $name, $spec)."\n"
+   ."    or Carp::croak(\"Attempted to access '${name}' but it is not set\");\n"
    ."  \$val;\n"
    ."}\n";
 }
@@ -655,11 +656,11 @@ sub _validate_codulatable {
   $invalid .= " $appended" if $appended;
 
   unless (ref $value and (ref $value eq 'CODE' or blessed($value))) {
-    die "$invalid or code-convertible object";
+    croak "$invalid or code-convertible object";
   }
 
   unless (eval { \&$value }) {
-    die "$invalid and could not be converted to a coderef: $@";
+    croak "$invalid and could not be converted to a coderef: $@";
   }
 
   1;
