@@ -1,24 +1,47 @@
 package Moo::Role;
 
 use Moo::_strictures;
-use Moo::_Utils;
+use Moo::_Utils qw(
+  _getglob
+  _getstash
+  _install_coderef
+  _install_modifier
+  _load_module
+  _name_coderef
+  _set_loaded
+  _unimport_coderefs
+);
+use Sub::Defer ();
+use Sub::Quote qw(quote_sub);
 use Role::Tiny ();
-our @ISA = qw(Role::Tiny);
+use Carp qw(croak);
+BEGIN { our @ISA = qw(Role::Tiny) }
+BEGIN {
+  our @CARP_NOT = qw(
+    Method::Generate::Accessor
+    Method::Generate::Constructor
+    Moo::sification
+    Moo::_Utils
+  );
+}
 
-our $VERSION = '2.000000';
+our $VERSION = '2.002_000';
 $VERSION = eval $VERSION;
 
 require Moo::sification;
+Moo::sification->import;
 
 BEGIN {
     *INFO = \%Role::Tiny::INFO;
     *APPLIED_TO = \%Role::Tiny::APPLIED_TO;
+    *COMPOSED = \%Role::Tiny::COMPOSED;
     *ON_ROLE_CREATE = \@Role::Tiny::ON_ROLE_CREATE;
 }
 
 our %INFO;
 our %APPLIED_TO;
 our %APPLY_DEFAULTS;
+our %COMPOSED;
 our @ON_ROLE_CREATE;
 
 sub _install_tracked {
@@ -29,23 +52,20 @@ sub _install_tracked {
 
 sub import {
   my $target = caller;
-  my ($me) = @_;
-
-  _set_loaded(caller);
-  strict->import;
-  warnings->import;
   if ($Moo::MAKERS{$target} and $Moo::MAKERS{$target}{is_class}) {
-    die "Cannot import Moo::Role into a Moo class";
+    croak "Cannot import Moo::Role into a Moo class";
   }
-  $INFO{$target} ||= {};
-  # get symbol table reference
-  my $stash = _getstash($target);
+  _set_loaded(caller);
+  goto &Role::Tiny::import;
+}
+
+sub _install_subs {
+  my ($me, $target) = @_;
   _install_tracked $target => has => sub {
     my $name_proto = shift;
     my @name_proto = ref $name_proto eq 'ARRAY' ? @$name_proto : $name_proto;
     if (@_ % 2 != 0) {
-      require Carp;
-      Carp::croak("Invalid options for " . join(', ', map "'$_'", @name_proto)
+      croak("Invalid options for " . join(', ', map "'$_'", @name_proto)
         . " attribute(s): even number of arguments expected, got " . scalar @_)
     }
     my %spec = @_;
@@ -62,7 +82,6 @@ sub import {
   # install before/after/around subs
   foreach my $type (qw(before after around)) {
     _install_tracked $target => $type => sub {
-      require Class::Method::Modifiers;
       push @{$INFO{$target}{modifiers}||=[]}, [ $type => @_ ];
       $me->_maybe_reset_handlemoose($target);
     };
@@ -75,25 +94,12 @@ sub import {
     $me->apply_roles_to_package($target, @_);
     $me->_maybe_reset_handlemoose($target);
   };
-  return if $me->is_role($target); # already exported into this package
-  $INFO{$target}{is_role} = 1;
   *{_getglob("${target}::meta")} = $me->can('meta');
-  # grab all *non-constant* (stash slot is not a scalarref) subs present
-  # in the symbol table and store their refaddrs (no need to forcibly
-  # inflate constant subs into real subs) - also add '' to here (this
-  # is used later) with a map to the coderefs in case of copying or re-use
-  my @not_methods = ('', map { *$_{CODE}||() } grep !ref($_), values %$stash);
-  @{$INFO{$target}{not_methods}={}}{@not_methods} = @not_methods;
-  # a role does itself
-  $APPLIED_TO{$target} = { $target => undef };
-
-  $_->($target)
-    for @ON_ROLE_CREATE;
 }
 
 push @ON_ROLE_CREATE, sub {
   my $target = shift;
-  if ($INC{'Moo/HandleMoose.pm'}) {
+  if ($INC{'Moo/HandleMoose.pm'} && !$Moo::sification::disabled) {
     Moo::HandleMoose::inject_fake_metaclass_for($target);
   }
 };
@@ -112,7 +118,7 @@ sub unimport {
 
 sub _maybe_reset_handlemoose {
   my ($class, $target) = @_;
-  if ($INC{"Moo/HandleMoose.pm"}) {
+  if ($INC{'Moo/HandleMoose.pm'} && !$Moo::sification::disabled) {
     Moo::HandleMoose::maybe_reinject_fake_metaclass_for($target);
   }
 }
@@ -121,7 +127,7 @@ sub methods_provided_by {
   my ($self, $role) = @_;
   _load_module($role);
   $self->_inhale_if_moose($role);
-  die "${role} is not a Moo::Role" unless $self->is_role($role);
+  croak "${role} is not a Moo::Role" unless $self->is_role($role);
   return $self->SUPER::methods_provided_by($role);
 }
 
@@ -176,10 +182,21 @@ sub _inhale_if_moose {
 
           my $tc = $get_constraint->($spec->{isa});
           my $check = $tc->_compiled_type_constraint;
+          my $tc_name = $tc->name;
+          $tc_name =~ s/([_\W])/sprintf('_%x', ord($1))/ge;
+          my $tc_var = "\$_check_for_${tc_name}";
 
-          $spec->{isa} = sub {
-            &$check or die "Type constraint failed for $_[0]"
-          };
+          $spec->{isa} = quote_sub
+            "${role}::${tc_name}",
+            qq{
+              &${tc_var} or Carp::croak "Type constraint failed for \$_[0]"
+            },
+            { $tc_var => \$check },
+            {
+              package => $role,
+              no_install => 1,
+            },
+          ;
 
           if ($spec->{coerce}) {
 
@@ -206,7 +223,6 @@ sub _inhale_if_moose {
         }
       }
     }
-    require Class::Method::Modifiers if @$mods;
     $INFO{$role}{inhaled_from_moose} = 1;
     $INFO{$role}{is_role} = 1;
   }
@@ -247,8 +263,13 @@ sub _make_accessors {
   }
 }
 
+sub _undefer_subs {
+  my ($self, $target, $role) = @_;
+  Sub::Defer::undefer_package($role);
+}
+
 sub role_application_steps {
-  qw(_handle_constructor _maybe_make_accessors),
+  qw(_handle_constructor _undefer_subs _maybe_make_accessors),
     $_[0]->SUPER::role_application_steps;
 }
 
@@ -257,7 +278,7 @@ sub apply_roles_to_package {
   foreach my $role (@roles) {
     _load_module($role);
     $me->_inhale_if_moose($role);
-    die "${role} is not a Moo::Role" unless $me->is_role($role);
+    croak "${role} is not a Moo::Role" unless $me->is_role($role);
   }
   $me->SUPER::apply_roles_to_package($to, @roles);
 }
@@ -266,7 +287,7 @@ sub apply_single_role_to_package {
   my ($me, $to, $role) = @_;
   _load_module($role);
   $me->_inhale_if_moose($role);
-  die "${role} is not a Moo::Role" unless $me->is_role($role);
+  croak "${role} is not a Moo::Role" unless $me->is_role($role);
   $me->SUPER::apply_single_role_to_package($to, $role);
 }
 
@@ -275,11 +296,12 @@ sub create_class_with_roles {
 
   my ($new_name, $compose_name) = $me->_composite_name($superclass, @roles);
 
-  return $new_name if $Role::Tiny::COMPOSED{class}{$new_name};
+  return $new_name if $COMPOSED{class}{$new_name};
 
   foreach my $role (@roles) {
-      _load_module($role);
-      $me->_inhale_if_moose($role);
+    _load_module($role);
+    $me->_inhale_if_moose($role);
+    croak "${role} is not a Moo::Role" unless $me->is_role($role);
   }
 
   my $m;
@@ -290,20 +312,16 @@ sub create_class_with_roles {
     *{_getglob("${new_name}::ISA")} = [ $superclass ];
     $Moo::MAKERS{$new_name} = {is_class => 1};
     $me->apply_roles_to_package($new_name, @roles);
-    _set_loaded($new_name, (caller)[1]);
-    return $new_name;
+  }
+  else {
+    $me->SUPER::create_class_with_roles($superclass, @roles);
+    $Moo::MAKERS{$new_name} = {is_class => 1};
+    $me->_handle_constructor($new_name, $_) for @roles;
   }
 
-  $me->SUPER::create_class_with_roles($superclass, @roles);
-
-  foreach my $role (@roles) {
-    die "${role} is not a Moo::Role" unless $me->is_role($role);
+  if ($INC{'Moo/HandleMoose.pm'} && !$Moo::sification::disabled) {
+    Moo::HandleMoose::inject_fake_metaclass_for($new_name);
   }
-
-  $Moo::MAKERS{$new_name} = {is_class => 1};
-
-  $me->_handle_constructor($new_name, $_) for @roles;
-
   _set_loaded($new_name, (caller)[1]);
   return $new_name;
 }
@@ -311,49 +329,68 @@ sub create_class_with_roles {
 sub apply_roles_to_object {
   my ($me, $object, @roles) = @_;
   my $new = $me->SUPER::apply_roles_to_object($object, @roles);
-  _set_loaded(ref $new, (caller)[1]);
+  my $class = ref $new;
+  _set_loaded($class, (caller)[1]);
 
-  my $apply_defaults = $APPLY_DEFAULTS{ref $new} ||= do {
+  my $apply_defaults = exists $APPLY_DEFAULTS{$class} ? $APPLY_DEFAULTS{$class}
+    : $APPLY_DEFAULTS{$class} = do {
     my %attrs = map { @{$INFO{$_}{attributes}||[]} } @roles;
 
     if ($INC{'Moo.pm'}
         and keys %attrs
-        and my $con_gen = Moo->_constructor_maker_for(ref $new)
-        and my $m = Moo->_accessor_maker_for(ref $new)) {
+        and my $con_gen = Moo->_constructor_maker_for($class)
+        and my $m = Moo->_accessor_maker_for($class)) {
       require Sub::Quote;
 
       my $specs = $con_gen->all_attribute_specs;
 
-      my $assign = "{no warnings 'void';\n";
       my %captures;
-      foreach my $name ( keys %attrs ) {
-        my $spec = $specs->{$name};
-        if ($m->has_eager_default($name, $spec)) {
-          my ($has, $has_cap)
-            = $m->generate_simple_has('$_[0]', $name, $spec);
-          my ($code, $pop_cap)
-            = $m->generate_use_default('$_[0]', $name, $spec, $has);
+      my $code = join('',
+        "no warnings 'void';\n",
+        ( map {
+          my $name = $_;
+          my $spec = $specs->{$name};
+          if ($m->has_eager_default($name, $spec)) {
+            my ($has, $has_cap)
+              = $m->generate_simple_has('$_[0]', $name, $spec);
+            my ($set, $pop_cap)
+              = $m->generate_use_default('$_[0]', $name, $spec, $has);
 
-          $assign .= $code . ";\n";
-          @captures{keys %$has_cap, keys %$pop_cap}
-            = (values %$has_cap, values %$pop_cap);
+            @captures{keys %$has_cap, keys %$pop_cap}
+              = (values %$has_cap, values %$pop_cap);
+            "($set),";
+          }
+          else {
+            ();
+          }
+        } sort keys %attrs ),
+      );
+      Sub::Quote::quote_sub(
+        "${class}::_apply_defaults",
+        $code,
+        \%captures,
+        {
+          package => $class,
+          no_install => 1,
         }
-      }
-      $assign .= "}";
-      Sub::Quote::quote_sub($assign, \%captures);
+      );
     }
     else {
-      sub {};
+      0;
     }
   };
-  $new->$apply_defaults;
+  if ($apply_defaults) {
+    local $Carp::Internal{+__PACKAGE__} = 1;
+    local $Carp::Internal{$class} = 1;
+    $new->$apply_defaults;
+  }
   return $new;
 }
 
 sub _composable_package_for {
   my ($self, $role) = @_;
   my $composed_name = 'Role::Tiny::_COMPOSABLE::'.$role;
-  return $composed_name if $Role::Tiny::COMPOSED{role}{$composed_name};
+  return $composed_name if $COMPOSED{role}{$composed_name};
   $self->_make_accessors_if_moose($composed_name, $role);
   $self->SUPER::_composable_package_for($role);
 }
@@ -379,6 +416,7 @@ sub does_role {
   if ($INC{'Moose.pm'}
       and $meta = Class::MOP::class_of($proto)
       and ref $meta ne 'Moo::HandleMoose::FakeMetaClass'
+      and $meta->can('does_role')
   ) {
     return $meta->does_role($role);
   }
@@ -419,34 +457,34 @@ Moo::Role - Minimal Object Orientation support for Roles
 
 =head1 SYNOPSIS
 
- package My::Role;
+  package My::Role;
 
- use Moo::Role;
- use strictures 2;
+  use Moo::Role;
+  use strictures 2;
 
- sub foo { ... }
+  sub foo { ... }
 
- sub bar { ... }
+  sub bar { ... }
 
- has baz => (
-   is => 'ro',
- );
+  has baz => (
+    is => 'ro',
+  );
 
- 1;
+  1;
 
 And elsewhere:
 
- package Some::Class;
+  package Some::Class;
 
- use Moo;
- use strictures 2;
+  use Moo;
+  use strictures 2;
 
- # bar gets imported, but not foo
- with('My::Role');
+  # bar gets imported, but not foo
+  with('My::Role');
 
- sub foo { ... }
+  sub foo { ... }
 
- 1;
+  1;
 
 =head1 DESCRIPTION
 
@@ -461,9 +499,9 @@ imported by this module.
 
 =head2 has
 
- has attr => (
-   is => 'ro',
- );
+  has attr => (
+    is => 'ro',
+  );
 
 Declares an attribute for the class to be composed into.  See
 L<Moo/has> for all options.
@@ -475,18 +513,18 @@ declared before the C<use Moo::Role> statement automatically.
 Anything imported after C<use Moo::Role> will be composed into
 consuming packages.  A package that consumes this role:
 
- package My::Role::ID;
+  package My::Role::ID;
 
- use Digest::MD5 qw(md5_hex);
- use Moo::Role;
- use Digest::SHA qw(sha1_hex);
+  use Digest::MD5 qw(md5_hex);
+  use Moo::Role;
+  use Digest::SHA qw(sha1_hex);
 
- requires 'name';
+  requires 'name';
 
- sub as_md5  { my ($self) = @_; return md5_hex($self->name);  }
- sub as_sha1 { my ($self) = @_; return sha1_hex($self->name); }
+  sub as_md5  { my ($self) = @_; return md5_hex($self->name);  }
+  sub as_sha1 { my ($self) = @_; return sha1_hex($self->name); }
 
- 1;
+  1;
 
 ..will now have a C<< $self->sha1_hex() >> method available to it
 that probably does not do what you expect.  On the other hand, a call
