@@ -5,14 +5,17 @@ sub _clean_eval { eval $_[0] }
 use Moo::_strictures;
 
 use Sub::Defer qw(defer_sub);
+use Moo::_Utils qw(_install_coderef);
 use Scalar::Util qw(weaken);
 use Exporter qw(import);
+use Carp qw(croak);
+BEGIN { our @CARP_NOT = qw(Sub::Defer) }
 use B ();
 BEGIN {
   *_HAVE_PERLSTRING = defined &B::perlstring ? sub(){1} : sub(){0};
 }
 
-our $VERSION = '2.000000';
+our $VERSION = '2.002_000';
 $VERSION = eval $VERSION;
 
 our @EXPORT = qw(quote_sub unquote_sub quoted_from_sub qsub);
@@ -21,7 +24,13 @@ our @EXPORT_OK = qw(quotify capture_unroll inlinify);
 our %QUOTED;
 
 sub quotify {
+  no warnings 'numeric';
   ! defined $_[0]     ? 'undef()'
+  # numeric detection
+  : (length( (my $dummy = '') & $_[0] )
+    && 0 + $_[0] eq $_[0]
+    && $_[0] * 0 == 0
+  ) ? $_[0]
   : _HAVE_PERLSTRING  ? B::perlstring($_[0])
   : qq["\Q$_[0]\E"];
 }
@@ -32,7 +41,7 @@ sub capture_unroll {
     '',
     map {
       /^([\@\%\$])/
-        or die "capture key should start with \@, \% or \$: $_";
+        or croak "capture key should start with \@, \% or \$: $_";
       (' ' x $indent).qq{my ${_} = ${1}{${from}->{${\quotify $_}}};\n};
     } keys %$captures
   );
@@ -75,7 +84,20 @@ sub quote_sub {
   undef($captures) if $captures && !keys %$captures;
   my $code = pop;
   my $name = $_[0];
-  my ($package, $hints, $bitmask, $hintshash) = (caller(0))[0,8,9,10];
+  if ($name) {
+    my $subname = $name;
+    my $package = $subname =~ s/(.*)::// ? $1 : caller;
+    $name = join '::', $package, $subname;
+    croak "package name $package too long!"
+      if length $package > 252;
+    croak "sub name $subname too long!"
+      if length $subname > 252;
+  }
+  my @caller = caller(0);
+  my $package   = $options->{package}      || $caller[0];
+  my $hints     = $options->{hints}        || $caller[8];
+  my $bitmask   = $options->{warning_bits} || $caller[9];
+  my $hintshash = $options->{hintshash}    || $caller[10];
   my $context
     ="# BEGIN quote_sub PRELUDE\n"
     ."package $package;\n"
@@ -90,17 +112,26 @@ sub quote_sub {
     ."}\n"
     ."# END quote_sub PRELUDE\n";
   $code = "$context$code";
-  my $quoted_info;
+
+  my $quoted_info = [ $name, $code, $captures ];
   my $unquoted;
-  my $deferred = defer_sub +($options->{no_install} ? undef : $name) => sub {
-    $unquoted if 0;
-    unquote_sub($quoted_info->[4]);
-  };
-  $quoted_info = [ $name, $code, $captures, \$unquoted, $deferred ];
-  weaken($quoted_info->[3]);
-  weaken($quoted_info->[4]);
-  weaken($QUOTED{$deferred} = $quoted_info);
-  return $deferred;
+  weaken($quoted_info->[3] = \$unquoted);
+  if ($options->{no_defer}) {
+    my $fake = \my $var;
+    local $QUOTED{$fake} = $quoted_info;
+    my $sub = unquote_sub($fake);
+    _install_coderef($name, $sub) if $name;
+    return $sub;
+  }
+  else {
+    my $deferred = defer_sub +($options->{no_install} ? undef : $name) => sub {
+      $unquoted if 0;
+      unquote_sub($quoted_info->[4]);
+    };
+    weaken($quoted_info->[4] = $deferred);
+    weaken($QUOTED{$deferred} = $quoted_info);
+    return $deferred;
+  }
 }
 
 sub quoted_from_sub {
@@ -121,6 +152,10 @@ sub unquote_sub {
   my $unquoted = $quoted->[3];
   unless ($unquoted && $$unquoted) {
     my ($name, $code, $captures) = @$quoted;
+    my $package;
+
+    ($package, $name) = $name =~ /(.*)::(.*)/
+      if $name;
 
     my $make_sub = "{\n";
 
@@ -134,11 +169,10 @@ sub unquote_sub {
           # disable the 'variable $x will not stay shared' warning since
           # we're not letting it escape from this scope anyway so there's
           # nothing trying to share it
-        ? "  no warnings 'closure';\n  sub ${name} {\n"
+        ? "  no warnings 'closure';\n  package ${package};\n  sub ${name} {\n"
         : "  \$\$_UNQUOTED = sub {\n"
     );
-    $make_sub .= "  \$_QUOTED if 0;\n";
-    $make_sub .= "  \$_UNQUOTED if 0;\n";
+    $make_sub .= "  (\$_QUOTED,\$_UNQUOTED) if 0;\n";
     $make_sub .= $code;
     $make_sub .= "  }".($name ? '' : ';')."\n";
     if ($name) {
@@ -148,7 +182,7 @@ sub unquote_sub {
     $ENV{SUB_QUOTE_DEBUG} && warn $make_sub;
     {
       no strict 'refs';
-      local *{$name} if $name;
+      local *{"${package}::${name}"} if $name;
       my ($success, $e);
       {
         local $@;
@@ -156,7 +190,7 @@ sub unquote_sub {
         $e = $@;
       }
       unless ($success) {
-        die "Eval went very, very wrong:\n\n${make_sub}\n\n$e";
+        croak "Eval went very, very wrong:\n\n${make_sub}\n\n$e";
       }
       weaken($QUOTED{$$unquoted} = $quoted);
     }
